
#include <cppuhelper/implbase1.hxx> // helper for implementations
#include <cppuhelper/implbase3.hxx> // helper for implementations
#include <cppuhelper/factory.hxx>
#include <cppuhelper/implementationentry.hxx>

#include <com/sun/star/sheet/XAddIn.hpp>
#include <com/sun/star/lang/XServiceName.hpp>
#include <com/sun/star/lang/XServiceInfo.hpp>
#include <org/openoffice/addin/Xepicsopenoffice.hpp>

#include <stdio.h>
#include <string.h>
#include <cstdlib>
#include <cmath>

#include <cadef.h>
#include <epicsThread.h>
#include <epicsMutex.h>
#include <epicsTime.h>

#define CA_PEND_IO_TIME  15.0  // seconds

#define NO_DATA_STRING       ((char *) "(n/a)")
#define BAD_FP_STRING        ((char *) "(NaN)")
#define UNKNOWN_TYPE_STRING  ((char *) "(unknown)")

//----------------------------------------------------------------------
using namespace ::rtl;
using namespace ::com::sun::star;
using namespace ::com::sun::star::uno;
using namespace ::com::sun::star::sheet;
//----------------------------------------------------------------------
// start our namespace
namespace _epicsopenoffice_impl_
{

// forward declaration
class epicsopenoffice_impl;

//----------------------------------------------------------------------
// define a simple class for keeping a dynamic list of pointers
//----------------------------------------------------------------------
typedef int (*POINTERLIST_COMPARE_PROC)(void *p1, void *p2);

class pointerList
{
  public:
    pointerList(int initialCount = 256, int extendCount = 256);
    virtual ~pointerList(void);
    void append(void *ptr);
    int count(void);
    void *itemAt(int idx);
    void *removeItemAt(int idx);
    void sort(POINTERLIST_COMPARE_PROC proc);

  protected:
    epicsMutexId  mutex;
    int           delta;
    int           currSize;
    int           currCount;
    void        **pPtrList;
};
//----------------------------------------------------------------------
// define an epicsThreadRunable descendant to do support stuff for us
//----------------------------------------------------------------------
class supportThread : public epicsThreadRunable
{
  public:
    supportThread(char *name, epicsopenoffice_impl *pImpl);
    virtual void run();
    void exit();

  private:
    bool exitFlag;
    epicsopenoffice_impl *pOurImpl;
    epicsThread thread;
};
//----------------------------------------------------------------------
// define a generic XVolatileResult descendant
//----------------------------------------------------------------------
class OurVolatileResult : public ::cppu::WeakImplHelper1< XVolatileResult >
{
  public:
    OurVolatileResult(Any *value_loc);
    virtual ~OurVolatileResult();
    void modified(void);

    // XVolatileResult interface
    virtual void SAL_CALL addResultListener( const Reference< XResultListener >& aListener ) throw (RuntimeException);
    virtual void SAL_CALL removeResultListener( const Reference< XResultListener >& aListener ) throw (RuntimeException);

  private:
    pointerList *ourListeners;
    Any *pAny;

    ResultEvent getResultEvent();
    void sendNotifies(void);
};
//----------------------------------------------------------------------
// define a class for a double data item
//----------------------------------------------------------------------
class DoubleItem
{
  public:
    DoubleItem(double value);
    virtual ~DoubleItem();

    Reference< XVolatileResult > getXVR(void);

    double get(void);
    void set(double D);

  private:
    double ourValue;
    Any ourAnyValue;
    OurVolatileResult *pXVR;
    Reference< XVolatileResult > ourXVR;
};
//----------------------------------------------------------------------
// define a class for an EPICS data item
//----------------------------------------------------------------------
class EpicsChannelItem
{
  public:
    EpicsChannelItem(OUString pvName);
    virtual ~EpicsChannelItem();

    Reference< XVolatileResult > getXVR(void);

    OUString getPvName(void);
    char *getAsciiPvName(void);
    bool process(bool publish = false);

    void connectionHandler(struct connection_handler_args ARGS);
    void monitorHandler(struct event_handler_args ARGS);

  private:
    OUString OU_pvName;
    char ASCII_pvName[128];
    OurVolatileResult *pXVR;
    Reference< XVolatileResult > ourXVR;
    Any ourValue;
    Any lastPublished;
    bool needSearch;
    chid channelID;
    evid eventID;
    int dataType;
};
//----------------------------------------------------------------------
// define our interface class
//----------------------------------------------------------------------
class epicsopenoffice_impl :
  public ::cppu::WeakImplHelper3< ::org::openoffice::addin::Xepicsopenoffice,
                                  lang::XServiceName,
                                  lang::XServiceInfo > //3-parameter template
{
  //Locale
  lang::Locale locale;

  public:
    epicsopenoffice_impl();
    ~epicsopenoffice_impl();

    bool process(void);

    // Xepicsopenoffice interface
    Reference< XVolatileResult > SAL_CALL EpicsGet(const OUString & name) throw (RuntimeException);
    Reference< XVolatileResult > SAL_CALL EpicsGetUpdateRate(void) throw (RuntimeException);
    Reference< XVolatileResult > SAL_CALL EpicsSetUpdateRate(double seconds) throw (RuntimeException);

    //XServiceName
    OUString SAL_CALL getServiceName(  ) throw (RuntimeException);

    //XServiceInfo
    OUString SAL_CALL getImplementationName(  ) throw (RuntimeException);
    ::sal_Bool SAL_CALL supportsService( const OUString& ServiceName ) throw (RuntimeException);
    Sequence< OUString > SAL_CALL getSupportedServiceNames(  ) throw (RuntimeException);

    //XLocalizable
    void SAL_CALL setLocale( const lang::Locale& eLocale ) throw (RuntimeException);
    lang::Locale SAL_CALL getLocale() throw (RuntimeException);

  private:
    pointerList *channelList;
    DoubleItem updateRate;
    supportThread ourThread;
    bool didCaInit;
    epicsTimeStamp lastUpdate;

    void doCaInit();
    EpicsChannelItem *findName(OUString name);
};
//----------------------------------------------------------------------
// internal routines
//----------------------------------------------------------------------
OUString createOUString(char *txt)
{
  return OUString(txt, strlen(txt), RTL_TEXTENCODING_ASCII_US);
}
//----------------------------------------------------------------------
static bool fp_bad(double num)
{
  // supposedly this works as a test for invalid floating-point numbers
  return(num != num);
}
/*--------------------------------------------------------------------*/
// EPICS callback routines
//----------------------------------------------------------------------
static void CALLBACK_EpicsConnFunc(struct connection_handler_args ARGS)
{
  try {
    EpicsChannelItem *pItem = (EpicsChannelItem *) ca_puser(ARGS.chid);
    if (pItem != NULL) pItem->connectionHandler(ARGS);
  } catch( ... ) {
  }
}
/*--------------------------------------------------------------------*/
static void CALLBACK_EpicsMonitorEventFunc(struct event_handler_args ARGS)
{
  try {
    EpicsChannelItem *pItem = (EpicsChannelItem *) ca_puser(ARGS.chid);
    if (pItem != NULL) pItem->monitorHandler(ARGS);
  } catch( ... ) {
  }
}
//----------------------------------------------------------------------
// class pointerList
//--------------------------------------------------------------------
pointerList::pointerList(int initialCount, int extendCount)
{
  mutex     = epicsMutexCreate();
  delta     = extendCount;
  currSize  = initialCount;
  currCount = 0;

  pPtrList = new void *[currSize];
  if (pPtrList) {
    for (int ii = 0; ii < currSize; ii++) pPtrList[ii] = NULL;
  }
}
//--------------------------------------------------------------------
pointerList::~pointerList(void)
{
  // we don't know how to call destructors for pointers we're storing
  // so we hope our owner has taken care of clearing us out, we just
  // dispose of the list of pointers we allocated
  if (pPtrList) {
    delete [] pPtrList;
    pPtrList = NULL;
  }

  // delete our mutex
  epicsMutexDestroy(mutex);
}
//--------------------------------------------------------------------
void pointerList::append(void *ptr)
{
  epicsMutexLock(mutex);

  if (currCount >= currSize) {
    // dynamic array is full, need to extend it
    int newSize = currSize + delta;
    void **newList = new void *[newSize];
    for (int ii = 0; ii < newSize; ii++) {
      newList[ii] = (ii < currCount) ? pPtrList[ii] : NULL;
    }
    currSize = newSize;
    delete [] pPtrList;
    pPtrList = newList;
  }

  // store pointer
  pPtrList[currCount] = ptr;
  currCount++;

  epicsMutexUnlock(mutex);
}
//--------------------------------------------------------------------
int pointerList::count(void)
{
  epicsMutexLock(mutex);
  int c = currCount;
  epicsMutexUnlock(mutex);
  return(c);
}
//--------------------------------------------------------------------
void *pointerList::itemAt(int idx)
{
  epicsMutexLock(mutex);
  void *retval = ((idx < 0) || (idx >= currCount)) ? NULL : pPtrList[idx];
  epicsMutexUnlock(mutex);
  return(retval);
}
//--------------------------------------------------------------------
void *pointerList::removeItemAt(int idx)
{
  epicsMutexLock(mutex);

  void *retval;
  if ((idx < 0) || (idx >= currCount))
    retval = NULL;
  else {
    retval = pPtrList[idx];
    for (int ii = idx; ii < currCount-1; ii++) {
      pPtrList[ii] = pPtrList[ii+1];
    }
    pPtrList[currCount-1] = NULL;
    currCount--;
  }

  epicsMutexUnlock(mutex);
  return(retval);
}
//--------------------------------------------------------------------
void pointerList::sort(POINTERLIST_COMPARE_PROC proc)
{
  // no compare procedure, don't sort
  if (! proc) return;

  epicsMutexLock(mutex);

  int gap = currCount / 2;
  do {
    int swaps;
    do {
      swaps = 0;
      for (int ii = 0; ii < (currCount - gap); ii++) {
        if (proc(pPtrList[ii],pPtrList[ii+gap]) > 0) {
          void *tmp = pPtrList[ii];
          pPtrList[ii] = pPtrList[ii+gap];
          pPtrList[ii+gap] = tmp;
          swaps++;
        }
      }
    } while (swaps != 0);
    gap = gap / 2;
  } while (gap != 0);

  epicsMutexUnlock(mutex);
}
//--------------------------------------------------------------------
// class OurVolatileResult
//--------------------------------------------------------------------
OurVolatileResult::OurVolatileResult(Any *value_loc)
{
  pAny = value_loc;
  ourListeners = new pointerList();
}
//--------------------------------------------------------------------
OurVolatileResult::~OurVolatileResult()
{
  if (ourListeners) {
    for (int ii = ourListeners->count()-1; ii >= 0; ii--) {
      Reference<XResultListener> *pObj = (Reference<XResultListener> *) ourListeners->itemAt(ii);
      if (! pObj) continue;
      delete pObj;
    }
    delete ourListeners;
    ourListeners = NULL;
  }
}
//--------------------------------------------------------------------
void OurVolatileResult::modified(void)
{
  sendNotifies();
}
//--------------------------------------------------------------------
void SAL_CALL OurVolatileResult::addResultListener( const Reference< XResultListener >& aListener ) throw (RuntimeException)
{
  // create a copy of this listener for our list
	Reference<XResultListener> *pObj = new Reference<XResultListener>(aListener);

  // add listener to our list
  ourListeners->append(pObj);

  // call acquire if this is our first listener
	if (ourListeners->count() == 1) acquire();

  // inform new listener of our current value
  (*pObj)->modified( getResultEvent() );
}
//--------------------------------------------------------------------
void SAL_CALL OurVolatileResult::removeResultListener( const Reference< XResultListener >& aListener ) throw (RuntimeException)
{
	acquire();

  // locate listener in our list
  for (int ii = 0; ii < ourListeners->count(); ii++) {
		Reference<XResultListener> *pObj = (Reference<XResultListener> *) ourListeners->itemAt(ii);
		if (*pObj == aListener) {
      // remove from list
      ourListeners->removeItemAt(ii);
      // delete object
      delete pObj;
      // call release if this was our last listener
      if (ourListeners->count() == 0) release();
			break;
		}
	}

	release();
}
//--------------------------------------------------------------------
ResultEvent OurVolatileResult::getResultEvent()
{
  ResultEvent aEvent( (cppu::OWeakObject*)this, (*pAny) );
  return( aEvent );
}
//----------------------------------------------------------------------
void OurVolatileResult::sendNotifies(void)
{
  ResultEvent aEvent = getResultEvent();
  for (int ii = 0; ii < ourListeners->count(); ii++) {
		Reference<XResultListener> *pObj = (Reference<XResultListener> *) ourListeners->itemAt(ii);
    if (! pObj) continue;
    (*pObj)->modified(aEvent);
  }
}
//----------------------------------------------------------------------
// class DoubleItem
//----------------------------------------------------------------------
DoubleItem::DoubleItem(double value)
   : ourValue(value)
{
  // store our starting value
  ourAnyValue <<= ourValue;

  // create our XVolatileResult item
  pXVR = new OurVolatileResult(&ourAnyValue);
  ourXVR = pXVR;
}
//----------------------------------------------------------------------
DoubleItem::~DoubleItem()
{
  // nothing to do here
}
//----------------------------------------------------------------------
Reference< XVolatileResult > DoubleItem::getXVR(void)
{
  return(ourXVR);
}
//----------------------------------------------------------------------
double DoubleItem::get(void)
{
  return(ourValue);
}
//----------------------------------------------------------------------
void DoubleItem::set(double D)
{
  ourValue = D;
  ourAnyValue <<= ourValue;
  pXVR->modified();
}
//----------------------------------------------------------------------
// implement EpicsChannelItem
//----------------------------------------------------------------------
EpicsChannelItem::EpicsChannelItem(OUString name)
{
  // create our XVolatileResult item
  pXVR = new OurVolatileResult(&ourValue);
  ourXVR = pXVR;

  // store pv name
  OU_pvName = name;

  // get ASCII version of name, for use in printfs and ca calls
  OString o = OUStringToOString(OU_pvName, RTL_TEXTENCODING_ASCII_US);
  strcpy(ASCII_pvName,o.pData->buffer);

  // init misc vars
  channelID = 0;
  eventID = 0;

  // initialize our values
  ourValue <<= createOUString(NO_DATA_STRING) ;
  lastPublished = ourValue;

  // flag need to search for our PV in process()
  needSearch = true;
}
//----------------------------------------------------------------------
EpicsChannelItem::~EpicsChannelItem()
{
  // clear monitoring, if we have any
  if (eventID) {
    ca_clear_subscription(eventID);
    eventID = 0;
  }

  // clear channel
  if (channelID) {
    ca_clear_channel(channelID);
    channelID = 0;
  }
}
//----------------------------------------------------------------------
Reference< XVolatileResult > EpicsChannelItem::getXVR(void)
{
  return(ourXVR);
}
//----------------------------------------------------------------------
void EpicsChannelItem::connectionHandler(struct connection_handler_args ARGS)
{
  // our PV's connection status has changed
  switch (ARGS.op) {
    // connection to PV is up
    case CA_OP_CONN_UP: {
        // determine data type of channel
        // request ENUMs as strings
        dataType = ca_field_type(ARGS.chid);
        if (dataType == DBF_ENUM) dataType = DBF_STRING;

        // cancel any previous monitor request
        if (eventID) {
          ca_clear_subscription(eventID);
          eventID = 0;
        }

        // request notification of value
        int status = ca_add_event(dataType,
                                  channelID,
                                  CALLBACK_EpicsMonitorEventFunc,
                                  this,
                                  &eventID);
        if (status != ECA_NORMAL) {
          // flag we've had an error
          char tmp[40];
          sprintf(tmp,"(ca err %d)",status);
          ourValue <<= createOUString(tmp);
          eventID = 0;
        }
    }
    break;

    // connection to PV is down
    case CA_OP_CONN_DOWN: {
       // flag we've got no data
       ourValue <<= createOUString(NO_DATA_STRING);
    }
    break;
  }
}
//----------------------------------------------------------------------
void EpicsChannelItem::monitorHandler(struct event_handler_args ARGS)
{
  if (ARGS.status != ECA_NORMAL) {
    // if got error, flag it
    char tmp[40];
    sprintf(tmp,"(ca err %d)",ARGS.status);
    ourValue <<= createOUString(tmp);
    return;
  }

  // got new data, fill in ourValue
  unsigned char *pData = (unsigned char *) ARGS.dbr;
  switch (dataType) {
    case DBF_STRING: {
      char *str = (char *) pData;
      ourValue <<= createOUString(str);
    }
    break;

    case DBF_SHORT: {
      short *s = (short *) pData;
      ourValue <<= (*s);
    }
    break;

    case DBF_FLOAT: {
      float *f = (float *) pData;
      if (fp_bad(*f))
        ourValue <<= createOUString(BAD_FP_STRING);
      else
        ourValue <<= (*f);
    }
    break;

    case DBF_CHAR: {
      char *c = (char *) pData;
      long L = (*c);
      ourValue <<= L;
    }
    break;

    case DBF_LONG: {
      long *l = (long *) pData;
      ourValue <<= (*l);
    }
    break;

    case DBF_DOUBLE: {
      double *d = (double *) pData;
      if (fp_bad(*d))
        ourValue <<= createOUString(BAD_FP_STRING);
      else
        ourValue <<= (*d);
    }
    break;

    default: {
      ourValue <<= createOUString(UNKNOWN_TYPE_STRING);
    }
  }
}
//----------------------------------------------------------------------
bool EpicsChannelItem::process(bool publish)
{
  // if need to search for our PV, do so
  if (needSearch) {
    // call CA to create channel (and start searching for it)
    int status = ca_create_channel(ASCII_pvName,
                                   CALLBACK_EpicsConnFunc,
                                   this,
                                   CA_PRIORITY_DEFAULT,
                                   &channelID);
    if (status != ECA_NORMAL) {
      // flag we've had an error
      char tmp[40];
      sprintf(tmp,"(ca err %d)",status);
      ourValue <<= createOUString(tmp);
      channelID = 0;
    }

    needSearch = false;
  }

  // if we're supposed to publish our value now, do so
  if (publish) {
    if (lastPublished != ourValue) {
      lastPublished = ourValue;
      pXVR->modified();
    }
  }

  return(true);
}
//----------------------------------------------------------------------
OUString EpicsChannelItem::getPvName(void)
{
  return(OU_pvName);
}
//----------------------------------------------------------------------
char *EpicsChannelItem::getAsciiPvName(void)
{
  return(ASCII_pvName);
}
//----------------------------------------------------------------------
// implement epicsopenoffice_impl
//----------------------------------------------------------------------
epicsopenoffice_impl::epicsopenoffice_impl()
   : didCaInit(false),
     updateRate(1.0),
     ourThread((char *)"supportThread",this)
{
  // create channel list
  channelList = new pointerList();

  // put current time in lastUpdate time stamp
  epicsTimeGetCurrent(&lastUpdate);
}
//----------------------------------------------------------------------
epicsopenoffice_impl::~epicsopenoffice_impl()
{
  // tell our thread it needs to exit
  ourThread.exit();

  // CA library termination
  ca_pend_io(1.0);
  ca_context_destroy();

  // destroy channels we created and channel list
  if (channelList) {
    for (int ii = channelList->count()-1; ii >= 0; ii--) {
      EpicsChannelItem *pItem = (EpicsChannelItem *) channelList->itemAt(ii);
      if (! pItem) continue;
      channelList->removeItemAt(ii);
      delete pItem;
    }
    delete channelList;
    channelList = NULL;
  }
}
//----------------------------------------------------------------------
void epicsopenoffice_impl::doCaInit()
{
  if (didCaInit) return;

  int status;
  char tmp[128];

  // initialize CA library, disabling preemptive callbacks (our callback
  // functions will then only be called from inside of other ca_xxx calls)
  status = ca_context_create(ca_disable_preemptive_callback);
  if (status != ECA_NORMAL) {
    sprintf(tmp,"ca_context_create failed with status %d",status);
    fprintf(stderr,"epicsreq error: %s\n",tmp);
    return;
  }

  // call pend i/o once
  status = ca_pend_io(CA_PEND_IO_TIME);
  if (status != ECA_NORMAL) {
    sprintf(tmp,"ca_pend_io failed with status %d",status);
    fprintf(stderr,"epicsreq error: %s\n",tmp);
    ca_context_destroy();
    return;
  }

  didCaInit = true;
}
//----------------------------------------------------------------------
bool epicsopenoffice_impl::process(void)
// called periodically by our support thread
// should return true if all is okay, false if support thread should abort
{
  // nothing to do if no channel list
  if (! channelList) return(false);

  // do our ca initialization stuff
  if (! didCaInit) doCaInit();

  // see if it's time to publish values, based on user's desired update rate
  epicsTimeStamp now;
  epicsTimeGetCurrent(&now);
  bool publish = false;
  if (epicsTimeDiffInSeconds(&now,&lastUpdate) >= updateRate.get()) {
    publish = true;
    epicsTimeGetCurrent(&lastUpdate);
  }

  // roll through all our EpicsChannelItem objects
  // we go backwards so we can delete items without messing up indexing
  for (int ii = channelList->count()-1; ii >= 0; ii--) {
    EpicsChannelItem *pItem = (EpicsChannelItem *) channelList->itemAt(ii);
    // let channel object do any necessary processing
    pItem->process(publish);
  }

  // do EPICS ca processing
  ca_pend_io(CA_PEND_IO_TIME);
  ca_poll();  // ca_poll calls ca_pend_event w/short timeout

  return(true);
}
//----------------------------------------------------------------------
EpicsChannelItem *epicsopenoffice_impl::findName(OUString name)
{
  // nothing to do if no channel list
  if (! channelList) return(NULL);

  // locate the EpicsChannelItem object with the given name
  // (not expected to be fast, shouldn't need to be called often...)
  EpicsChannelItem *pItem = NULL;
  for (int ii = 0; ii < channelList->count(); ii++) {
    EpicsChannelItem *pLook = (EpicsChannelItem *) channelList->itemAt(ii);
    if (name == pLook->getPvName()) {
      pItem = pLook;
      break;
    }
  }
  return(pItem);
}
//----------------------------------------------------------------------
Reference<XVolatileResult> SAL_CALL epicsopenoffice_impl::EpicsGetUpdateRate(void) throw (RuntimeException)
{
  return(updateRate.getXVR());
}
//----------------------------------------------------------------------
Reference<XVolatileResult> SAL_CALL epicsopenoffice_impl::EpicsSetUpdateRate(double seconds) throw (RuntimeException)
{
  updateRate.set( (seconds < 0.1) ? 0.1 : seconds );
  return( EpicsGetUpdateRate() );
}
//----------------------------------------------------------------------
Reference<XVolatileResult> SAL_CALL epicsopenoffice_impl::EpicsGet(const OUString & name) throw (RuntimeException)
{
  // nothing to do if no channel list
  if (! channelList) return(NULL);

  // see if we already have an EpicsChannelItem item for this name
  // create one (and add it to our list) if we don't
  EpicsChannelItem *pItem = findName(name);
  if (! pItem) {
    pItem = new EpicsChannelItem(name);
    channelList->append(pItem);
  }

  // return item to caller
  return(pItem->getXVR());
}
//----------------------------------------------------------------------
#define _serviceName_ ((char *) "org.openoffice.addin.epicsopenoffice")
OUString epicsopenoffice_impl::getServiceName(  ) throw (RuntimeException)
{
  return createOUString(_serviceName_);
}
//----------------------------------------------------------------------
static OUString getImplementationName_epicsopenoffice_impl() throw (RuntimeException)
{
  return OUString(RTL_CONSTASCII_USTRINGPARAM("org.openoffice.addin.epicsopenoffice_impl.epicsopenoffice"));
}
//----------------------------------------------------------------------
OUString epicsopenoffice_impl::getImplementationName() throw (RuntimeException)
{
  return getImplementationName_epicsopenoffice_impl();
}
//----------------------------------------------------------------------
::sal_Bool epicsopenoffice_impl::supportsService( OUString const & serviceName ) throw (RuntimeException)
{
  return serviceName.equalsAsciiL(_serviceName_, strlen(_serviceName_)-1);
}
//----------------------------------------------------------------------
static Sequence< OUString > getSupportedServiceNames_epicsopenoffice_impl() throw (RuntimeException)
{
  Sequence< OUString > name(1);
  name[0] = createOUString(_serviceName_);
  return name;
}
//----------------------------------------------------------------------
Sequence< OUString > epicsopenoffice_impl::getSupportedServiceNames() throw (RuntimeException)
{
  return getSupportedServiceNames_epicsopenoffice_impl();
}
//----------------------------------------------------------------------
void epicsopenoffice_impl::setLocale( const lang::Locale& eLocale ) throw (RuntimeException)
{
  locale = eLocale;
}
//----------------------------------------------------------------------
lang::Locale epicsopenoffice_impl::getLocale(  ) throw (RuntimeException)
{
  return locale;
}
//----------------------------------------------------------------------
static Reference< XInterface > SAL_CALL create_epicsopenoffice_impl( Reference< XComponentContext > const & xContext ) SAL_THROW( () )
{
  return static_cast< ::cppu::OWeakObject * > ( new epicsopenoffice_impl() );
}
//----------------------------------------------------------------------
static struct ::cppu::ImplementationEntry s_component_entries[] =
{
  { create_epicsopenoffice_impl, getImplementationName_epicsopenoffice_impl,
    getSupportedServiceNames_epicsopenoffice_impl, ::cppu::createSingleComponentFactory, 0, 0}
  ,
  { 0, 0, 0, 0, 0, 0}
};
//----------------------------------------------------------------------
// class supportThread
//----------------------------------------------------------------------
supportThread::supportThread(char *name, epicsopenoffice_impl *pImpl)
   : thread(*this,name,epicsThreadGetStackSize(epicsThreadStackSmall),epicsThreadPriorityScanLow)
{
  pOurImpl = pImpl;
  thread.start();
}
//----------------------------------------------------------------------
void supportThread::exit()
{
  epicsTimeStamp start;
  epicsTimeGetCurrent(&start);
  exitFlag = true;
  while (exitFlag) {
    epicsThreadSleep(0.1);
    epicsTimeStamp now;
    epicsTimeGetCurrent(&now);
    if (epicsTimeDiffInSeconds(&now,&start) >= 3.0) break;
  }
}
//----------------------------------------------------------------------
void supportThread::run()
{
  exitFlag = false;
  while (! exitFlag) {
    if (! pOurImpl->process()) break;
    thread.sleep(0.1);
  }
  exitFlag = false;
}
//----------------------------------------------------------------------
// end of our namespace
}
//----------------------------------------------------------------------
// externally-callable functions required
//----------------------------------------------------------------------
extern "C"
{
  void SAL_CALL component_getImplementationEnvironment( sal_Char const ** ppEnvTypeName, uno_Environment ** ppEnv )
  {
    *ppEnvTypeName = CPPU_CURRENT_LANGUAGE_BINDING_NAME;
  }

  sal_Bool SAL_CALL component_writeInfo( lang::XMultiServiceFactory * xMgr, registry::XRegistryKey * xRegistry )
  {
    return ::cppu::component_writeInfoHelper( xMgr, xRegistry, ::_epicsopenoffice_impl_::s_component_entries );
  }

  void * SAL_CALL component_getFactory( sal_Char const * implName,
                                        lang::XMultiServiceFactory * xMgr,
                                        registry::XRegistryKey * xRegistry )
  {
    return ::cppu::component_getFactoryHelper(implName, xMgr, xRegistry,
                                              ::_epicsopenoffice_impl_::s_component_entries );
  }
}
//----------------------------------------------------------------------

